@@ -530,8 +530,12 @@ orbit_small_demarshal (CORBA_Object           obj,
 	    GIOP_LOCATION_FORWARD) {
 		
 		*cnx = ORBit_handle_location_forward (recv_buffer, obj);
-		tprintf (" Exception: forward");
-
+		tprintf (" Exception: forward (%p)", *cnx);
+		if (!*cnx) {
+			CORBA_exception_set_system(ev, ex_CORBA_MARSHAL,
+						   CORBA_COMPLETED_MAYBE);
+			return MARSHAL_SYS_EXCEPTION_INCOMPLETE;
+		}
 		return MARSHAL_RETRY;
 	} else {
 		ORBit_handle_exception_array (
@@ -586,6 +590,7 @@ ORBit_small_invoke_stub (CORBA_Object       obj,
 	ORBit_OAObject          adaptor_obj;
 	GIOPRecvBuffer         *recv_buffer = NULL;
 	CORBA_Object            xt_proxy = CORBA_OBJECT_NIL;
+	ORBitPolicy            *invoke_policy = CORBA_OBJECT_NIL;
 
 	if (!obj) {
 		dprintf (MESSAGES, "Cannot invoke method on null object\n");
@@ -594,11 +599,15 @@ ORBit_small_invoke_stub (CORBA_Object       obj,
 		goto clean_out;
 	}
 
+	if ((invoke_policy = ORBit_object_get_policy (obj)))
+		ORBit_policy_push (invoke_policy);
+
 	adaptor_obj = obj->adaptor_obj;
 
 	if (adaptor_obj) {
 		/* FIXME: unchecked cast */
-		if (ORBit_poa_allow_cross_thread_call ((ORBit_POAObject) adaptor_obj)) {
+		if (ORBit_poa_allow_cross_thread_call ((ORBit_POAObject) adaptor_obj,
+						       m_data->flags)) {
 			tprintf_header (obj, m_data);
 			tprintf ("[in-proc]");
 			ORBit_small_handle_request (adaptor_obj, m_data->name, ret,
@@ -606,6 +615,11 @@ ORBit_small_invoke_stub (CORBA_Object       obj,
 			goto clean_out;
 		} else {
 			tprintf ("[in-proc-XT]");
+			/*
+			 * FIXME: this is _really_ slow, can easily be optimised
+			 * by shoving the GIOP data straight on the incoming
+			 * queue
+			 */
 			xt_proxy = ORBit_objref_get_proxy (obj);
 			obj = xt_proxy;
 		}
@@ -638,7 +652,6 @@ ORBit_small_invoke_stub (CORBA_Object       obj,
 		goto clean_out;
 	}
 
-	giop_recv_buffer_unuse (recv_buffer);
 	recv_buffer = giop_recv_buffer_get (&mqe);
 
 	switch (orbit_small_demarshal (obj, &cnx, recv_buffer, ev,
@@ -673,6 +686,10 @@ ORBit_small_invoke_stub (CORBA_Object       obj,
 	tprintf_end_method ();
 	if (cnx)
 		giop_connection_unref (cnx);
+	if (invoke_policy) {
+		ORBit_policy_pop ();
+		ORBit_policy_unref (invoke_policy);
+	}
 	return;
 
  system_exception:
